from fastapi import APIRouter, Depends, HTTPException, status
from sqlmodel import select
from sqlmodel.ext.asyncio.session import AsyncSession

from lexy.api.deps import get_current_active_user, get_current_active_superuser, get_db
from lexy.models.user import User, UserCreate, UserRead, UserUpdate


router = APIRouter()


@router.get(
    "/users",
    response_model=list[UserRead],
    status_code=status.HTTP_200_OK,
    name="get_users",
    description="Get all users (superuser only)",
    dependencies=[Depends(get_current_active_superuser)],
)
async def get_users(session: AsyncSession = Depends(get_db)) -> list[UserRead]:
    result = await session.exec(select(User))
    users = result.all()
    return users


@router.get(
    "/users/me",
    response_model=UserRead,
    status_code=status.HTTP_200_OK,
    name="get_user_me",
    description="Get current user",
)
async def get_user_me(
    current_user: User = Depends(get_current_active_user),
) -> UserRead:
    # TODO: change the next line to `return current_user` after updating SQLModel
    return UserRead.model_validate(current_user)


@router.post(
    "/users",
    response_model=UserRead,
    status_code=status.HTTP_201_CREATED,
    name="create_user",
    description="Create a new user (superuser only)",
    dependencies=[Depends(get_current_active_superuser)],
)
async def create_user(
    user: UserCreate, session: AsyncSession = Depends(get_db)
) -> UserRead:
    # check if user already exists
    result = await session.exec(select(User).where(User.email == user.email))
    existing_user = result.first()
    if existing_user:
        raise HTTPException(status_code=400, detail="User email is already registered")
    # create new user
    new_user = User.create(**user.model_dump())
    session.add(new_user)
    await session.commit()
    await session.refresh(new_user)
    return new_user


@router.get(
    "/users/{user_id}",
    response_model=UserRead,
    status_code=status.HTTP_200_OK,
    name="get_user",
    description="Get a specific user (superuser only)",
    dependencies=[Depends(get_current_active_superuser)],
)
async def get_user(user_id: int, session: AsyncSession = Depends(get_db)) -> UserRead:
    user = await session.get(User, user_id)
    if user is None:
        raise HTTPException(status_code=404, detail="User not found")
    # TODO: change the next line to `return user` after updating SQLModel
    return UserRead.model_validate(user)


@router.delete(
    "/users/{user_id}",
    status_code=status.HTTP_200_OK,
    name="delete_user",
    description="Delete a specific user (superuser only)",
    dependencies=[Depends(get_current_active_superuser)],
)
async def delete_user(
    user_id: int, session: AsyncSession = Depends(get_db)
) -> dict[str, str]:
    user = await session.get(User, user_id)
    if user is None:
        raise HTTPException(status_code=404, detail="User not found")
    await session.delete(user)
    await session.commit()
    return {"detail": "User deleted", "user_id": user_id}


@router.patch(
    "/users/{user_id}",
    response_model=UserRead,
    status_code=status.HTTP_200_OK,
    name="update_user",
    description="Update a specific user (superuser only)",
    dependencies=[Depends(get_current_active_superuser)],
)
async def update_user(
    user_id: int, user: UserUpdate, session: AsyncSession = Depends(get_db)
) -> UserRead:
    db_user = await session.get(User, user_id)
    if db_user is None:
        raise HTTPException(status_code=404, detail="User not found")
    update_data = user.model_dump(exclude_unset=True)
    for field in update_data:
        setattr(db_user, field, update_data[field])
    await session.commit()
    await session.refresh(db_user)
    # TODO: change the next line to `return db_user` after updating SQLModel
    return UserRead.model_validate(db_user)



from lexy.models.document import DocumentBase
from lexy.transformers import lexy_transformer

import torch
from sentence_transformers import SentenceTransformer


torch.set_num_threads(1)
model = SentenceTransformer("sentence-transformers/all-MiniLM-L6-v2")


@lexy_transformer(name="text.embeddings.minilm")
def text_embeddings(
    sentences: list[str | DocumentBase] | str | DocumentBase,
) -> torch.Tensor:
    """Embed sentences using SentenceTransformer.

    Args:
        sentences: A single sentence or a list of sentences to embed. Each sentence can
            be either a string or a DocumentBase instance.

    Returns:
        torch.Tensor: The embeddings of the provided sentences.
    """
    if isinstance(sentences, DocumentBase):
        sentences = sentences.content
    elif isinstance(sentences, list):
        sentences = [s.content if isinstance(s, DocumentBase) else s for s in sentences]
    return model.encode(sentences, batch_size=len(sentences))



import os
import shutil

import click
from dotenv import dotenv_values, set_key


@click.group()
@click.option("--debug", is_flag=True, help="Debug mode")
@click.pass_context
def cli(ctx, debug):
    ctx.ensure_object(dict)
    ctx.obj["debug"] = debug


@cli.command(help="Initialize Lexy")
@click.option(
    "--project-dir",
    default=".",
    help="Project home directory",
    show_default=True,
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, writable=True, resolve_path=True
    ),
)
@click.option(
    "--pipeline-dir",
    default="./pipelines",
    help="Directory to store Lexy pipelines",
    show_default=True,
    type=click.Path(
        exists=False, file_okay=False, dir_okay=True, writable=True, resolve_path=True
    ),
)
@click.option("--overwrite", is_flag=True, help="Overwrite existing values")
@click.option("--no-input", is_flag=True, help="Do not prompt for input")
@click.pass_context
def init(ctx, project_dir, pipeline_dir, overwrite, no_input):
    if ctx.obj["debug"]:
        click.echo(f"Debug mode: {ctx.obj['debug']}")

    # Project directory
    default_project_dir = os.path.expanduser(project_dir)
    if not no_input:
        # Prompt user
        project_dir = click.prompt("Project directory", default=default_project_dir)
    else:
        # Use default values
        project_dir = default_project_dir
    # Check if the project directory exists
    if not os.path.exists(project_dir):
        click.secho(f"Directory does not exist: {project_dir}", err=True, fg="red")
        raise click.Abort()

    # Pipeline directory
    default_pipeline_dir = os.path.expanduser(pipeline_dir)
    if not no_input:
        # Prompt user
        pipeline_dir = click.prompt(
            "Lexy pipeline directory", default=default_pipeline_dir
        )
    else:
        # Use default values
        pipeline_dir = default_pipeline_dir

    # Check if an .env file exists in the project directory
    env_file = os.path.join(project_dir, ".env")
    if not os.path.exists(env_file):
        # Copy the .env.example template to the current working directory
        env_example_file = os.path.join(
            os.path.dirname(__file__), "templates", ".env.example"
        )
        shutil.copy(env_example_file, env_file)
        click.secho(f"Copied .env.example to {env_file}", fg="blue")
    else:
        click.secho(f"Found existing .env file: {env_file}", fg="blue")

    # Load the .env file
    dotenv_config = dotenv_values(env_file)
    if ctx.obj["debug"]:
        click.secho(f"Loaded .env file: {env_file}", fg="blue")

    # Check if PIPELINE_DIR is already set in the .env file
    if (
        "PIPELINE_DIR" in dotenv_config
        and dotenv_config["PIPELINE_DIR"] == pipeline_dir
    ):
        click.secho(f"PIPELINE_DIR already set: {pipeline_dir}", fg="blue")
    elif (
        "PIPELINE_DIR" in dotenv_config
        and dotenv_config["PIPELINE_DIR"] != pipeline_dir
    ):
        click.secho(
            f"PIPELINE_DIR already set: {dotenv_config['PIPELINE_DIR']}",
            err=True,
            fg="yellow",
        )
        if not no_input:
            click.confirm("Do you want to overwrite the existing value?", abort=True)
            set_key(env_file, "PIPELINE_DIR", pipeline_dir)
            click.secho(f"Set PIPELINE_DIR in .env file: {pipeline_dir}", fg="blue")
        else:
            if overwrite:
                set_key(env_file, "PIPELINE_DIR", pipeline_dir)
                click.secho(f"Set PIPELINE_DIR in .env file: {pipeline_dir}", fg="blue")
            else:
                click.secho(
                    "Existing value for PIPELINE_DIR. Use --overwrite to overwrite.",
                    err=True,
                    fg="red",
                )
                raise click.Abort()
    else:
        # Set PIPELINE_DIR in the .env file
        set_key(env_file, "PIPELINE_DIR", pipeline_dir)
        click.secho(f"Set PIPELINE_DIR in .env file: {pipeline_dir}", fg="blue")

    # Create the pipeline directory if it doesn't exist
    if not os.path.exists(pipeline_dir):
        os.makedirs(pipeline_dir)
        click.secho(f"Created pipeline directory: {pipeline_dir}", fg="blue")
    else:
        click.secho(f"Found existing pipeline directory: {pipeline_dir}", fg="blue")

    # Copy the pipeline template to the pipeline directory
    template_dir = os.path.join(os.path.dirname(__file__), "templates", "pipelines")
    files_to_copy = [
        f
        for f in os.listdir(template_dir)
        if os.path.isfile(os.path.join(template_dir, f))
    ]
    for file in files_to_copy:
        # if file exists, log a warning and skip
        if os.path.exists(os.path.join(pipeline_dir, file)):
            existing_file = os.path.join(pipeline_dir, file)
            click.secho(f"File already exists: {existing_file}", err=True, fg="yellow")
            continue
        else:
            shutil.copy(os.path.join(template_dir, file), pipeline_dir)
            if ctx.obj["debug"]:
                click.secho(f"Copied {file} to {pipeline_dir}", fg="blue")

    click.secho("Finished initializing Lexy", fg="green")


@cli.command(help="Create docker-compose file")
@click.option(
    "--file",
    default="docker-compose.yaml",
    help="Filepath to save docker-compose file",
    show_default=True,
    type=click.Path(
        exists=False, file_okay=True, dir_okay=False, writable=True, resolve_path=True
    ),
)
@click.option("--overwrite", is_flag=True, help="Overwrite existing file")
@click.option("--no-input", is_flag=True, help="Do not prompt for input")
@click.pass_context
def docker(ctx, file, overwrite, no_input):
    if ctx.obj["debug"]:
        click.echo(f"Debug mode: {ctx.obj['debug']}")

    # Expand the user's home directory
    default_file = file

    if not no_input:
        # Prompt user
        file = click.prompt("Docker compose file", default=file)
    else:
        # Use default values
        file = default_file

    # File to copy
    template_file = os.path.join(
        os.path.dirname(__file__), "templates", "docker-compose.yaml"
    )

    # Check if the file already exists
    if os.path.exists(file):
        click.secho(f"File already exists: {file}", err=True, fg="yellow")
        if not no_input:
            click.confirm("Do you want to overwrite the existing file?", abort=True)
            os.remove(file)
            click.secho(f"Removed existing file: {file}", fg="blue")
        else:
            if overwrite:
                os.remove(file)
                click.secho(f"Removed existing file: {file}", fg="blue")
            else:
                click.secho(
                    f"Existing file: {file}. Use --overwrite to overwrite ",
                    err=True,
                    fg="red",
                )
                raise click.Abort()

    # Copy the template file to the current working directory
    shutil.copy(template_file, file)
    click.secho(f"Copied docker-compose.yaml to {file}", fg="green")


@cli.group(help="Run tests", invoke_without_command=True)
@click.pass_context
def test(ctx):
    if ctx.invoked_subcommand is None:
        click.echo("Running all tests")


@test.command(help="Run tests for Lexy server")
def server():
    click.echo("Running: pytest -s lexy_tests")


@test.command(help="Run tests for Lexy client")
def client():
    click.echo("Running: pytest -s sdk-python")


if __name__ == "__main__":
    cli(obj={})



import datetime
import logging

from google.cloud import storage
from google.oauth2 import service_account
from google.auth.exceptions import DefaultCredentialsError

from lexy.core.config import settings
from lexy.storage.base import StorageClient

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class GoogleCredentialsError(Exception):
    """Raised when missing credentials file."""

    pass


class GCSClient(StorageClient):
    def __init__(
        self, credentials_file: str = settings.GOOGLE_APPLICATION_CREDENTIALS, **kwargs
    ):
        if credentials_file:
            logger.info(
                f"Creating GCS client using credentials file: {credentials_file}"
            )
            credentials = service_account.Credentials.from_service_account_file(
                credentials_file
            )
        else:
            raise GoogleCredentialsError(
                "Missing credentials file for Google Cloud Storage. Make sure to set "
                "`GOOGLE_APPLICATION_CREDENTIALS` to a valid JSON file containing "
                "service account credentials."
            )
        self.client = storage.Client(credentials=credentials, **kwargs)

    def is_authenticated(self) -> bool:
        try:
            self.client.list_buckets(max_results=1)
            return True
        except DefaultCredentialsError:
            return False
        except Exception:
            raise

    def list_buckets(self) -> list[str]:
        buckets = self.client.list_buckets()
        return [bucket.name for bucket in buckets]

    def upload_object(
        self, fileobj, bucket_name: str, object_name: str, rewind: bool = True
    ) -> dict:
        bucket = self.client.bucket(bucket_name)
        blob: storage.blob.Blob = bucket.blob(object_name)
        if isinstance(fileobj, str):
            blob.upload_from_filename(fileobj)
        else:
            blob.upload_from_file(fileobj, rewind=rewind)
        return {
            "storage_service": "gcs",
            "storage_bucket": bucket_name,
            "storage_key": object_name,
            # "storage_url": f"https://storage.googleapis.com/{bucket_name}/{object_name}",
            # "storage_uri": f"gs://{bucket_name}/{object_name}",
        }

    def generate_presigned_url(
        self, bucket_name: str, object_name: str, expiration: int = 3600
    ) -> str:
        bucket = self.client.bucket(bucket_name)
        blob: storage.blob.Blob = bucket.blob(object_name)
        url = blob.generate_signed_url(
            expiration=datetime.timedelta(seconds=expiration), version="v4"
        )
        return url

    def delete_object(self, bucket_name: str, object_name: str) -> None:
        bucket = self.client.bucket(bucket_name)
        blob: storage.blob.Blob = bucket.blob(object_name)
        blob.delete()



default_data = {
    "collections": [
        {"collection_name": "default", "description": "Default collection"},
        {"collection_name": "code", "description": "Github code repos"},
    ],
    "transformers": [
        {
            "transformer_id": "text.embeddings.minilm",
            "path": "lexy.transformers.embeddings.text_embeddings",
            "description": 'Text embeddings using "sentence-transformers/all-MiniLM-L6-v2"',
        },
        {
            "transformer_id": "text.embeddings.clip",
            "path": "lexy.transformers.multimodal.text_embeddings_clip",
            "description": 'Text embeddings using "openai/clip-vit-base-patch32"',
        },
        {
            "transformer_id": "image.embeddings.clip",
            "path": "lexy.transformers.multimodal.image_embeddings_clip",
            "description": 'Image embeddings using "openai/clip-vit-base-patch32"',
        },
        {
            "transformer_id": "text.embeddings.openai",
            "path": "lexy.transformers.openai.text_embeddings",
            "description": "Text embeddings using the OpenAI API",
        },
        {
            "transformer_id": "text.embeddings.openai-ada-002",
            "path": "lexy.transformers.openai.text_embeddings_ada_002",
            "description": 'Text embeddings using OpenAI\'s "text-embedding-ada-002" model',
        },
        {
            "transformer_id": "text.embeddings.openai-3-small",
            "path": "lexy.transformers.openai.text_embeddings_3_small",
            "description": 'Text embeddings using OpenAI\'s "text-embedding-3-small" model',
        },
        {
            "transformer_id": "text.embeddings.openai-3-large",
            "path": "lexy.transformers.openai.text_embeddings_3_large",
            "description": 'Text embeddings using OpenAI\'s "text-embedding-3-large" model',
        },
        {
            "transformer_id": "text.counter.word_counter",
            "path": "lexy.transformers.counter.word_counter",
            "description": "Returns count of words and the longest word",
        },
    ],
    "indexes": [
        {
            "index_id": "default_text_embeddings",
            "description": "Text embeddings for default collection",
            "index_table_schema": {
                "title": "DefaultTextEmbeddings",
                "type": "object",
                "properties": {
                    "document_id": {
                        "title": "Document Id",
                        "type": "string",
                        "format": "uuid",
                    },
                    "embedding": {
                        "title": "Embedding",
                        "type": "array",
                        "items": {"type": "number"},
                    },
                    "text": {"title": "Text", "type": "string"},
                    "meta": {"title": "Meta", "default": {}, "type": "object"},
                    "custom_id": {"title": "Custom Id", "type": "string"},
                    "embedding_id": {
                        "title": "Embedding Id",
                        "type": "string",
                        "format": "uuid",
                    },
                    "created_at": {
                        "title": "Created At",
                        "type": "string",
                        "format": "date-time",
                    },
                    "updated_at": {
                        "title": "Updated At",
                        "type": "string",
                        "format": "date-time",
                    },
                    "task_id": {"title": "Task Id", "type": "string", "format": "uuid"},
                },
                "required": ["document_id", "embedding", "created_at", "updated_at"],
            },
            "index_fields": {
                "embedding": {
                    "type": "embedding",
                    "extras": {"dims": 384, "model": "text.embeddings.minilm"},
                },
                "text": {"type": "string", "optional": True},
            },
        }
    ],
    "bindings": [
        {
            "collection_name": "default",
            "transformer_id": "text.embeddings.minilm",
            "index_id": "default_text_embeddings",
            "description": "Default binding",
            "execution_params": {},
            "transformer_params": {"lexy_index_fields": ["embedding"]},
            "filters": {},
            "status": "on",
        }
    ],
}

sample_docs = {
    "default_collection_sample_docs": [
        {
            "content": "This is my first document! It's great!",
            "collection_name": "default",
        },
        {
            "content": "Starlink is a satellite internet constellation operated by American aerospace company SpaceX, "
            "providing coverage to over 60 countries.",
            "collection_name": "default",
        },
        {
            "content": "A latent space is an embedding of a set of items within a manifold in which items resembling "
            "each other are positioned closer to one another.",
            "collection_name": "default",
        },
    ],
    "code_collection_sample_docs": [
        {
            "collection_name": "code",
            "content": "import this",
            "meta": {
                "filename": "main.py",
                "file_extension": "py",
                "language": "python",
            },
        },
        {
            "collection_name": "code",
            "content": "def multiply(a, b):"
            "    return a * b"
            ""
            ""
            "if __name__ == '__main__':"
            "    print(multiply(2, 3))",
            "meta": {
                "filename": "multiply.py",
                "file_extension": "py",
                "language": "python",
            },
        },
    ],
}



from io import BytesIO
from typing import Union

from fastapi import APIRouter, Depends, HTTPException, Query, status, UploadFile
from PIL import Image
from sqlmodel import delete, exists, select
from sqlmodel.ext.asyncio.session import AsyncSession

from lexy import crud
from lexy.db.session import get_session
from lexy.storage.client import (
    construct_key_for_document,
    construct_key_for_thumbnail,
    get_storage_client,
    StorageClient,
)
from lexy.models.collection import Collection, CollectionCreate, CollectionUpdate
from lexy.models.document import Document, DocumentCreate
from lexy.core.config import settings
from lexy.core.events import generate_tasks_for_document


router = APIRouter()


@router.get(
    "/collections",
    response_model=Union[Collection, list[Collection]],
    status_code=status.HTTP_200_OK,
    name="get_collections",
    description="Get collections",
)
async def get_collections(
    collection_name: str | None = None, session: AsyncSession = Depends(get_session)
) -> Collection | list[Collection]:
    if collection_name:
        collection = await crud.get_collection_by_name(
            session=session, collection_name=collection_name
        )
        if not collection:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail="Collection not found"
            )
        return collection
    result = await session.exec(select(Collection))
    collections = result.all()
    return collections


@router.post(
    "/collections",
    response_model=Collection,
    status_code=status.HTTP_201_CREATED,
    name="create_collection",
    description="Create a new collection",
)
async def create_collection(
    collection: CollectionCreate, session: AsyncSession = Depends(get_session)
) -> Collection:
    # check if collection with that name already exists
    existing_collection = await crud.get_collection_by_name(
        session=session, collection_name=collection.collection_name
    )
    if existing_collection:
        raise HTTPException(
            status_code=400, detail="Collection with that name already exists"
        )

    db_collection = Collection.model_validate(collection)
    session.add(db_collection)
    await session.commit()
    await session.refresh(db_collection)
    return db_collection


@router.delete(
    "/collections",
    status_code=status.HTTP_200_OK,
    name="delete_collection_by_name",
    description="Delete a collection by name",
)
async def delete_collection_by_name(
    collection_name: str,
    delete_documents: bool = False,
    session: AsyncSession = Depends(get_session),
) -> dict:
    # get the collection
    collection = await crud.get_collection_by_name(
        session=session, collection_name=collection_name
    )
    if not collection:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Collection not found"
        )
    collection_id = collection.collection_id

    # delete any documents in the collection
    deleted_count = 0
    if delete_documents is True:
        statement = delete(Document).where(Document.collection_id == collection_id)
        result = await session.exec(statement)
        deleted_count = result.rowcount
    else:
        # check if there are any related documents
        statement = select(
            exists().where(Document.collection_id == collection_id)
        ).select_from(Document)
        result = await session.exec(statement)
        has_documents = result.first()
        if has_documents:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail="There are still documents in this collection. "
                "Set delete_documents=True to delete them.",
            )

    # delete the collection
    await session.delete(collection)
    await session.commit()
    return {
        "msg": "Collection deleted",
        "collection_id": collection_id,
        "documents_deleted": deleted_count,
    }


@router.get(
    "/collections/{collection_id}",
    response_model=Collection,
    status_code=status.HTTP_200_OK,
    name="get_collection",
    description="Get a collection by ID",
)
async def get_collection(
    collection_id: str, session: AsyncSession = Depends(get_session)
) -> Collection:
    collection = await crud.get_collection_by_id(
        session=session, collection_id=collection_id
    )
    if not collection:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Collection not found"
        )
    return collection


@router.patch(
    "/collections/{collection_id}",
    response_model=Collection,
    status_code=status.HTTP_200_OK,
    name="update_collection",
    description="Update a collection",
)
async def update_collection(
    collection_id: str,
    collection: CollectionUpdate,
    session: AsyncSession = Depends(get_session),
) -> Collection:
    db_collection = await crud.get_collection_by_id(
        session=session, collection_id=collection_id
    )
    if not db_collection:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Collection not found"
        )
    collection_data = collection.model_dump(exclude_unset=True)
    # if collection_name is being updated, check if a collection with that name already
    # exists
    if "collection_name" in collection_data:
        existing_collection = await crud.get_collection_by_name(
            session=session, collection_name=collection_data["collection_name"]
        )
        if existing_collection:
            raise HTTPException(
                status_code=400, detail="Collection with that name already exists"
            )
    for key, value in collection_data.items():
        setattr(db_collection, key, value)
    session.add(db_collection)
    await session.commit()
    await session.refresh(db_collection)
    return db_collection


@router.delete(
    "/collections/{collection_id}",
    status_code=status.HTTP_200_OK,
    name="delete_collection",
    description="Delete a collection",
)
async def delete_collection(
    collection_id: str,
    delete_documents: bool = False,
    session: AsyncSession = Depends(get_session),
) -> dict:
    # get the collection
    collection = await crud.get_collection_by_id(
        session=session, collection_id=collection_id
    )
    if not collection:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Collection not found"
        )

    # delete any documents in the collection
    deleted_count = 0
    if delete_documents is True:
        statement = delete(Document).where(Document.collection_id == collection_id)
        result = await session.exec(statement)
        deleted_count = result.rowcount
    else:
        # check if there are any related documents
        statement = select(
            exists().where(Document.collection_id == collection_id)
        ).select_from(Document)
        result = await session.exec(statement)
        has_documents = result.first()
        if has_documents:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail="There are still documents in this collection. "
                "Set delete_documents=True to delete them.",
            )

    # delete the collection
    await session.delete(collection)
    await session.commit()
    return {
        "msg": "Collection deleted",
        "collection_id": collection_id,
        "documents_deleted": deleted_count,
    }


@router.get(
    "/collections/{collection_id}/documents",
    response_model=list[Document],
    status_code=status.HTTP_200_OK,
    name="Get documents in collection",
    description="Get documents in a collection",
)
async def get_collection_documents(
    collection_id: str,
    limit: int = Query(100, gt=0, le=1000),
    offset: int = 0,
    session: AsyncSession = Depends(get_session),
) -> list[Document]:
    # get the collection
    collection = await crud.get_collection_by_id(
        session=session, collection_id=collection_id
    )
    if not collection:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Collection not found"
        )
    # get documents in the collection
    result = await session.exec(
        select(Document)
        .where(Document.collection_id == collection_id)
        .limit(limit)
        .offset(offset)
    )
    documents = result.all()
    return documents


@router.post(
    "/collections/{collection_id}/documents",
    status_code=status.HTTP_201_CREATED,
    name="Add documents to collection",
    description="Add documents to a collection",
)
async def add_collection_documents(
    collection_id: str,
    documents: list[DocumentCreate],
    session: AsyncSession = Depends(get_session),
) -> list[dict]:
    # get the collection
    collection = await crud.get_collection_by_id(
        session=session, collection_id=collection_id
    )
    if not collection:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Collection not found"
        )
    # add documents
    docs_added = []
    for doc in documents:
        # Check if a document_id is provided and if it already exists
        if doc.document_id:
            existing_doc = await session.get(Document, doc.document_id)
            if existing_doc:
                raise HTTPException(
                    status_code=400,
                    detail={
                        "msg": f"A document with this ID already exists: "
                        f"{doc.document_id}.",
                        "document_id": str(doc.document_id),
                    },
                )
        document = Document(**doc.model_dump(), collection_id=collection.collection_id)
        session.add(document)
        await session.commit()
        await session.refresh(document)
        tasks = await generate_tasks_for_document(document)
        docs_added.append({"document": document, "tasks": tasks})
    return docs_added


# TODO: refactor logic to combine with `lexy.api.endpoints.documents.upload_documents`
@router.post(
    "/collections/{collection_id}/documents/upload",
    status_code=status.HTTP_201_CREATED,
    name="Upload documents to collection",
    description="Upload documents to a collection",
)
async def upload_collection_documents(
    collection_id: str,
    files: list[UploadFile],
    session: AsyncSession = Depends(get_session),
    storage_client: StorageClient = Depends(get_storage_client),
) -> list[dict]:
    # get the collection
    collection = await crud.get_collection_by_id(
        session=session, collection_id=collection_id
    )
    if not collection:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Collection not found"
        )
    # TODO: get storage client from collection config
    # storage_client = get_storage_client(service=collection.config.get('storage_service', None))
    if collection.config.get("store_files") and not storage_client:
        storage_service = collection.config.get("storage_service", None)
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"Storage client not configured (service: {storage_service})",
        )
    storage_bucket = collection.config.get(
        "storage_bucket", settings.DEFAULT_STORAGE_BUCKET
    )
    if collection.config.get("store_files") and not storage_bucket:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Storage bucket not configured for this collection",
        )
    storage_prefix = collection.config.get("storage_prefix", None)

    docs_uploaded = []

    for file in files:
        file_dict = {
            "content_type": file.content_type,
            "filename": file.filename,
            "size": file.size,
        }

        # TODO: replace file.filename with the unique document id - will require saving
        #  the document in the DB first but for now, we're using the filename as the
        #  document_id, which is equivalent to the following:
        #    document_key = f"collections/{collection_id}/documents/{file.filename}"
        # uncomment the following line when ready
        # document_key = await construct_key_for_document(
        #     document=document,
        #     path_prefix=storage_prefix,
        #     filename=file.filename
        # )
        document_key = await construct_key_for_document(
            collection_id=collection_id,
            document_id=file.filename,
            path_prefix=storage_prefix,
        )

        # TODO: move this to a separate parsing function - don't read into memory if
        #  not parsing
        file_content = await file.read()
        file_in_memory = BytesIO(file_content)
        file_in_memory.seek(0)

        # "pre-processing" based on file type
        if file.content_type.startswith("image/"):
            # logic for image files
            file_dict["type"] = "image"
            doc_content = f"<Image({file.filename})>"
        elif file.content_type.startswith("text/"):
            # logic for text files
            file_dict["type"] = "text"
            doc_content = file_content.decode("utf-8")
        elif file.content_type.startswith("application/pdf"):
            # logic for pdf files
            file_dict["type"] = "pdf"
            doc_content = f"<PDF({file.filename})>"
        elif (
            file.content_type
            == "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
        ):
            # logic for docx files
            file_dict["type"] = "docx"
            doc_content = f"<DOCX({file.filename})>"
        elif file.content_type.startswith("video/"):
            # logic for video files
            file_dict["type"] = "video"
            doc_content = f"<Video({file.filename})>"
        else:
            # logic for all other files
            file_dict["type"] = "file"
            doc_content = f"<File({file.filename})>"

        # check for an existing doc based on content
        docs = await crud.get_documents_by_collection_id_and_content(
            session=session, collection_id=collection_id, content=doc_content
        )
        if docs:
            file_dict["document"] = docs[0]
            docs_uploaded.append(file_dict)
            continue

        # "post-processing" based on file type
        if file.content_type.startswith("image/"):
            # all we're doing here is getting the width and height - if we're
            # generating thumbnails, we'll use this instance as the input
            img = Image.open(file_in_memory)
            width, height = img.size
            file_dict["image"] = {"width": width, "height": height}

            # generate thumbnails
            if collection.config.get("generate_thumbnails") and collection.config.get(
                "store_files"
            ):
                # generate thumbnails and upload to storage
                thumbnails = {}
                for dims in settings.IMAGE_THUMBNAIL_SIZES:
                    thumbnail_key = await construct_key_for_thumbnail(
                        dims=dims,
                        collection_id=collection_id,
                        document_id=file.filename,
                        path_prefix=storage_prefix,
                    )
                    img_copy = img.copy()
                    img_copy.thumbnail(dims, Image.Resampling.LANCZOS)
                    img_byte_arr = BytesIO()
                    img_copy.save(img_byte_arr, format=img.format or "JPEG")
                    # Note: img_byte_arr.seek(0) is run in storage_client.upload_object
                    # by default `rewind=True`
                    storage_thumbnail_meta = storage_client.upload_object(
                        img_byte_arr, storage_bucket, thumbnail_key
                    )
                    thumbnails[f"{dims[0]}x{dims[1]}"] = storage_thumbnail_meta
                file_dict["image"]["thumbnails"] = thumbnails
        elif (
            file.content_type
            == "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
        ):
            # # example processing for docx
            # import docx
            # doc = docx.Document(BytesIO(file.file.read()))
            # num_paragraphs = len(doc.paragraphs)
            # num_tables = len(doc.tables)
            pass

        # Upload the file to storage
        if collection.config.get("store_files"):
            storage_document_meta = storage_client.upload_object(
                file_in_memory, storage_bucket, document_key
            )
            file_dict.update(storage_document_meta)

        document = Document(
            content=doc_content, meta=file_dict, collection_id=collection_id
        )
        session.add(document)
        await session.commit()
        await session.refresh(document)

        # generate tasks
        tasks = await generate_tasks_for_document(
            document, storage_client=storage_client
        )
        file_dict["document"] = document
        file_dict["tasks"] = tasks

        docs_uploaded.append(file_dict)

    return docs_uploaded


# TODO: Implement bulk update
# @router.patch("/collections/{collection_id}/documents",
#               status_code=status.HTTP_200_OK,
#               name="Bulk update documents in collection",
#               description="Update all documents in a collection")
# async def update_collection_documents(collection_id: str,
#                                       document_update: DocumentUpdate,
#                                       # documents: list[DocumentUpdate],
#                                       session: AsyncSession = Depends(get_session)) -> dict:
#     pass


@router.delete(
    "/collections/{collection_id}/documents",
    status_code=status.HTTP_200_OK,
    name="Bulk delete documents in collection",
    description="Delete all documents in a collection",
)
async def delete_collection_documents(
    collection_id: str, session: AsyncSession = Depends(get_session)
) -> dict:
    # get the collection
    collection = await crud.get_collection_by_id(
        session=session, collection_id=collection_id
    )
    if not collection:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Collection not found"
        )
    # delete the documents
    statement = delete(Document).where(Document.collection_id == collection_id)
    result = await session.exec(statement)
    deleted_count = result.rowcount
    await session.commit()
    return {
        "msg": "Documents deleted",
        "collection_id": collection_id,
        "documents_deleted": deleted_count,
    }



from datetime import datetime
from typing import Any, Optional, TYPE_CHECKING
from uuid import uuid4

from sqlalchemy import Column, DateTime, func, String
from sqlalchemy.dialects.postgresql import JSONB
from sqlmodel import SQLModel, Field, Relationship

from lexy.core.config import settings

if TYPE_CHECKING:
    from lexy.models.binding import Binding
    from lexy.models.document import Document


def generate_short_uid() -> str:
    return uuid4().hex[:8]


class CollectionBase(SQLModel):
    collection_name: str = Field(
        unique=True,
        min_length=1,
        max_length=56,
        # Postgres identifiers are limited to 63 characters (so 63 - len('zzcol__') = 56)
        # TODO: switch back to `regex=` (or `pattern=`) once SQLModel bug is fixed
        #   https://github.com/tiangolo/sqlmodel/discussions/735
        schema_extra={"pattern": r"^[a-z_][a-z0-9_]{0,55}$"},
    )
    description: Optional[str] = None
    config: Optional[dict[str, Any]] = Field(
        sa_column=Column(JSONB), default=settings.COLLECTION_DEFAULT_CONFIG
    )


class Collection(CollectionBase, table=True):
    __tablename__ = "collections"
    collection_id: str = Field(
        default=None,
        sa_column=Column(
            String(length=8), default=generate_short_uid, primary_key=True, unique=True
        ),
    )
    created_at: datetime = Field(
        default=None,
        sa_column=Column(
            DateTime(timezone=True), nullable=False, server_default=func.now()
        ),
    )
    updated_at: datetime = Field(
        default=None,
        sa_column=Column(
            DateTime(timezone=True),
            nullable=False,
            server_default=func.now(),
            onupdate=func.now(),
        ),
    )
    documents: list["Document"] = Relationship(
        back_populates="collection", sa_relationship_kwargs={"lazy": "select"}
    )
    bindings: list["Binding"] = Relationship(
        back_populates="collection",
        sa_relationship_kwargs={"lazy": "selectin", "cascade": "all, delete-orphan"},
    )


class CollectionCreate(CollectionBase):
    pass


class CollectionUpdate(CollectionBase):
    collection_name: Optional[str] = Field(
        default=None,
        unique=True,
        min_length=1,
        max_length=56,
        # Postgres identifiers are limited to 63 characters (so 63 - len('zzcol__') = 56)
        # TODO: switch back to `regex=` (or `pattern=`) once SQLModel bug is fixed
        #   https://github.com/tiangolo/sqlmodel/discussions/735
        schema_extra={"pattern": r"^[a-z_][a-z0-9_]{0,55}$"},
    )
    description: Optional[str] = None
    config: Optional[dict[str, Any]] = None



import pytest
from sqlmodel import select

from lexy.models.index import Index, IndexCreate


class TestIndex:
    def test_hello(self):
        assert True

    @pytest.mark.asyncio
    async def test_get_indexes(self, async_session):
        result = await async_session.exec(select(Index))
        indexes = result.all()
        assert len(indexes) == 1
        index_ids = [index.index_id for index in indexes]
        assert "default_text_embeddings" in index_ids

    def test_get_indexes_with_client(self, client):
        response = client.get(
            "/api/indexes",
        )
        assert response.status_code == 200, response.text
        data = response.json()
        assert len(data) == 1
        index_ids = [index["index_id"] for index in data]
        assert "default_text_embeddings" in index_ids

    @pytest.mark.asyncio
    async def test_get_indexes_with_async_client(self, async_client):
        response = await async_client.get(
            "/api/indexes",
        )
        assert response.status_code == 200, response.text
        data = response.json()
        assert len(data) == 1
        index_ids = [index["index_id"] for index in data]
        assert "default_text_embeddings" in index_ids

    @pytest.mark.asyncio
    async def test_get_index(self, async_session):
        result = await async_session.exec(
            select(Index).where(Index.index_id == "default_text_embeddings")
        )
        index = result.first()
        assert index.index_id == "default_text_embeddings"
        assert index.description == "Text embeddings for default collection"
        assert set(index.index_fields.keys()) == {"text", "embedding"}

    @pytest.mark.asyncio
    async def test_create_index_row_only(self, async_session):
        # define index fields
        test_index_fields = {
            "text": {"type": "text"},
            "embedding": {
                "type": "embedding",
                "extras": {"dims": 384, "model": "text.embeddings.minilm"},
            },
            "meta": {"type": "object"},
        }
        index = IndexCreate(
            index_id="test_index",
            description="Test Index",
            index_fields=test_index_fields,
        )

        db_index = Index.model_validate(index)
        async_session.add(db_index)
        await async_session.commit()
        await async_session.refresh(db_index)
        assert db_index.index_id == "test_index"
        assert db_index.description == "Test Index"

        result = await async_session.exec(
            select(Index).where(Index.index_id == "test_index")
        )
        indexes = result.all()
        assert len(indexes) == 1
        test_index = indexes[0]
        assert test_index.index_id == "test_index"
        assert test_index.description == "Test Index"
        assert set(test_index.index_fields.keys()) == {"text", "embedding", "meta"}

    @pytest.mark.asyncio
    async def test_delete_index_row_only(self, async_session):
        result = await async_session.exec(
            select(Index).where(Index.index_id == "test_index")
        )
        index = result.first()
        assert index is not None

        await async_session.delete(index)
        await async_session.commit()

        result = await async_session.exec(
            select(Index).where(Index.index_id == "test_index")
        )
        index = result.first()
        assert index is None


class TestIndexManager:
    # TODO: create fixture for IndexManager and run all relevant tests
    #   - test creating index model
    #   - test creating index table
    #   - test saving index record(s)
    #   - test getting index record(s)
    #   - test updating index record(s)
    #   - test deleting index record(s)
    #   - test dropping index table
    pass


class TestIndexModel:
    def test_create_index(self):
        index = IndexCreate(index_id="test_index")
        assert index.index_id == "test_index"
        index = IndexCreate(index_id="_te5t")
        assert index.index_id == "_te5t"
        index = IndexCreate(index_id="_myindex" * 7)
        assert index.index_id == "_myindex" * 7

    def test_create_index_with_invalid_id(self):
        with pytest.raises(ValueError):
            IndexCreate(index_id="", description="Test Index")  # blank
        with pytest.raises(ValueError):
            IndexCreate(index_id="test index", description="Test Index")  # space
        with pytest.raises(ValueError):
            IndexCreate(index_id="test-index", description="Test Index")  # hyphen
        with pytest.raises(ValueError):
            IndexCreate(index_id="Test", description="Test Index")  # uppercase
        with pytest.raises(ValueError):
            IndexCreate(index_id="1abc", description="Test Index")  # starts with number
        with pytest.raises(ValueError):
            IndexCreate(index_id="_myindex" * 8, description="Test Index")  # too long



import logging
import os
import time

from fastapi import APIRouter, status

from lexy.core.config import settings
from lexy.core.celery_app import get_task_info
from lexy.core.events import celery, restart_celery_worker


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
router = APIRouter()


@router.get("/", status_code=status.HTTP_200_OK, name="Root")
async def root():
    return {"Say": "Hello!"}


@router.get("/ping", status_code=status.HTTP_200_OK, name="Ping")
async def pong():
    return {"ping": "pong!"}


@router.get("/lexy-conf", status_code=status.HTTP_200_OK, name="Lexy settings")
async def lexy_settings():
    return settings.model_dump()


@router.get("/celery-conf", status_code=status.HTTP_200_OK, name="Celery Conf")
async def celery_conf(with_defaults: bool = False):
    return {"Celery conf": celery.control.inspect().conf(with_defaults=with_defaults)}


@router.get("/celery-status", status_code=status.HTTP_200_OK, name="Celery Status")
async def celery_status():
    return {"Celery status": celery.control.inspect().stats()}


@router.get(
    "/celery-registered-tasks",
    status_code=status.HTTP_200_OK,
    name="Celery Registered Tasks",
)
async def celery_registered_tasks():
    return {
        "Celery registered tasks": celery.control.inspect().registered_tasks(),
        "celery.tasks.keys()": sorted(celery.tasks.keys()),
    }


@router.get("/celery-active", status_code=status.HTTP_200_OK, name="Celery Active")
async def celery_active():
    return {"Celery active": celery.control.inspect().active()}


@router.get(
    "/celery-restart-worker",
    status_code=status.HTTP_200_OK,
    name="Celery Restart Worker",
)
async def celery_restart_worker(worker_name: str = "celery@celeryworker"):
    celery_restart_response = restart_celery_worker(worker_name)
    logger.info(f"Response: {celery_restart_response}")
    time.sleep(0.5)
    return {"Celery restart worker": celery_restart_response}


@router.get(
    "/tasks/{task_id}",
    response_model=dict,
    status_code=status.HTTP_200_OK,
    name="get_task_status",
    description="Get task status",
)
async def get_task_status(task_id: str, verbose: bool = False) -> dict:
    return get_task_info(task_id, verbose=verbose)


@router.get(
    "/tasks",
    response_model=dict,
    status_code=status.HTTP_200_OK,
    name="get_all_tasks",
    description="Get all tasks",
)
async def get_all_tasks() -> dict:
    return {
        "active": celery.control.inspect().active(),
        "scheduled": celery.control.inspect().scheduled(),
        "reserved": celery.control.inspect().reserved(),
        "revoked": celery.control.inspect().revoked(),
        "stats": celery.control.inspect().stats(),
        "registered_tasks": celery.control.inspect().registered_tasks(),
        "conf": celery.control.inspect().conf(with_defaults=False),
    }


@router.get("/index-manager", status_code=status.HTTP_200_OK, name="get_index_manager")
async def get_index_manager() -> dict:
    from lexy.api.deps import index_manager

    index_models_and_tables = dict(
        (model, index_manager.index_models[model].__table__.name)
        for model in index_manager.index_models
    )
    return {"index_models_and_tables": index_models_and_tables}


@router.get(
    "/index-manager-celery",
    status_code=status.HTTP_200_OK,
    name="get_index_manager_from_celery",
)
async def get_index_manager_from_celery() -> dict:
    from lexy.core.celery_tasks import index_manager

    index_models_and_tables = dict(
        (model, index_manager.index_models[model].__table__.name)
        for model in index_manager.index_models
    )
    return {"index_models_and_tables": index_models_and_tables}


@router.get("/os-env", status_code=status.HTTP_200_OK, name="os_env")
async def os_env():
    return dict(os.environ)



import functools

from celery import shared_task


def lexy_fields(func):
    """A decorator to assign index field labels to a transformer's return values.

    Args:
        func: The transformer function to decorate

    Returns:
        function: The decorated transformer function

    Examples:
        >>> @lexy_fields
        ... def add_and_subtract(a, b):
        ...     return a + b, a - b
        ...
        >>> add_and_subtract(5, 3, lexy_fields=["sum", "difference"])
        {'sum': 8, 'difference': 2}
        >>> add_and_subtract(5, 3, lexy_fields=["addition", "subtraction"])
        {'addition': 8, 'subtraction': 2}
        >>> add_and_subtract(5, 3)
        (8, 2)
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        fields = kwargs.pop("lexy_fields", [])
        results = func(*args, **kwargs)

        # if no lexy fields are provided, return the original results
        if not fields:
            return results

        # if the function returns a single value, wrap it in a tuple
        if not isinstance(results, tuple):
            results = (results,)

        # check if the number of lexy fields matches the number of results
        if len(fields) != len(results):
            raise ValueError(
                f"Expected {len(fields)} lexy fields ({', '.join(fields)}), "
                f"but got {len(results)} return values from '{func.__name__}'."
            )

        # zip together labels and results and return as a dictionary inside a list
        return [dict(zip(fields, results))]

    # modify the docstring to include documentation for the "lexy_fields" parameter
    lexy_doc = (
        "\n\n"
        "Lexy Transformer options:\n\n"
        "\tlexy_fields: list, optional\n"
        "\t    A list of field names to be used for the return values.\n"
    )
    if func.__doc__:
        wrapper.__doc__ = func.__doc__ + lexy_doc
    else:
        wrapper.__doc__ = lexy_doc

    return wrapper


def lexy_transformer(name: str, **task_kwargs):
    """A decorator to register a transformer function with Lexy.

    Args:
        name: The name of the transformer. The task will be registered with Celery as
            "lexy.transformers.{name}"
        **task_kwargs: Keyword arguments to pass to the Celery task decorator
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            lexy_index_fields = kwargs.pop("lexy_index_fields", [])
            results = func(*args, **kwargs)

            # if no index fields are provided, return the original results
            if not lexy_index_fields:
                return results

            # if the function returns a single value, wrap it in a tuple
            if not isinstance(results, tuple):
                results = (results,)

            # check if the number of index fields matches the number of results
            if len(lexy_index_fields) != len(results):
                raise ValueError(
                    f"Expected {len(lexy_index_fields)} index fields "
                    f"({', '.join(lexy_index_fields)}), but got {len(results)} return "
                    f"values from '{func.__name__}'."
                )

            # zip together labels and results and return as a dictionary inside a list
            return [dict(zip(lexy_index_fields, results))]

        # Modify the docstring to include documentation for the "lexy_index_fields"
        # parameter
        lexy_doc = (
            "\n\n"
            "Lexy Transformer options:\n\n"
            "\tlexy_index_fields: list, optional\n"
            "\t    A list of index fields to be used for the return values.\n"
        )
        if func.__doc__:
            wrapper.__doc__ = func.__doc__ + lexy_doc
        else:
            wrapper.__doc__ = lexy_doc

        wrapper = shared_task(name=f"lexy.transformers.{name}", **task_kwargs)(wrapper)
        return wrapper

    return decorator



from datetime import datetime
from typing import Optional, TYPE_CHECKING

from sqlalchemy import Column, DateTime, func, String
from sqlmodel import SQLModel, Field, Relationship

if TYPE_CHECKING:
    from sqlalchemy.engine.default import DefaultExecutionContext
    from lexy.models.binding import Binding


def default_celery_task_name(transformer_id: str) -> str:
    return f"lexy.transformers.{transformer_id}"


def sa_default_celery_task_name(context: "DefaultExecutionContext") -> str:
    transformer_id = context.get_current_parameters()["transformer_id"]
    return default_celery_task_name(transformer_id)


class TransformerBase(SQLModel):
    transformer_id: str = Field(
        default=None,
        primary_key=True,
        min_length=1,
        max_length=255,
        # TODO: switch back to `regex=` (or `pattern=`) once SQLModel bug is fixed
        #   https://github.com/tiangolo/sqlmodel/discussions/735
        schema_extra={"pattern": r"^[a-zA-Z][a-zA-Z0-9_.-]+$"},
    )
    path: Optional[str] = Field(
        default=None,
        min_length=1,
        max_length=255,
        # TODO: switch back to `regex=` (or `pattern=`) once SQLModel bug is fixed
        #   https://github.com/tiangolo/sqlmodel/discussions/735
        schema_extra={"pattern": r"^[a-zA-Z][a-zA-Z0-9_.]+$"},
    )
    description: Optional[str] = None


class Transformer(TransformerBase, table=True):
    __tablename__ = "transformers"
    created_at: datetime = Field(
        default=None,
        sa_column=Column(
            DateTime(timezone=True), nullable=False, server_default=func.now()
        ),
    )
    updated_at: datetime = Field(
        default=None,
        sa_column=Column(
            DateTime(timezone=True),
            nullable=False,
            server_default=func.now(),
            onupdate=func.now(),
        ),
    )
    bindings: list["Binding"] = Relationship(back_populates="transformer")
    celery_task_name: Optional[str] = Field(
        default=None,
        sa_column=Column(
            String,
            default=sa_default_celery_task_name,
            nullable=True,
            unique=True,
        ),
    )


class TransformerCreate(TransformerBase):
    pass


class TransformerUpdate(TransformerBase):
    description: Optional[str] = None
    path: Optional[str] = None



from datetime import datetime
from enum import Enum
from typing import Any, Optional

from pydantic import model_validator
from sqlalchemy import Column, DateTime, ForeignKey, func
from sqlalchemy.dialects.postgresql import JSONB
from sqlmodel import SQLModel, Field, Relationship

from lexy.models.collection import Collection
from lexy.models.index import Index
from lexy.models.transformer import Transformer
from lexy.schemas.filters import Filter


class BindingStatus(str, Enum):
    PENDING = "pending"
    ON = "on"
    OFF = "off"
    DETACHED = "detached"


# Shared properties
class BindingBase(SQLModel):
    collection_id: Optional[str] = None
    transformer_id: str = Field(
        foreign_key="transformers.transformer_id", default=None, nullable=True
    )
    index_id: str = Field(
        sa_column_args=(ForeignKey("indexes.index_id", ondelete="CASCADE"),),
        default=None,
        nullable=True,
    )
    description: Optional[str] = None
    execution_params: dict[str, Any] = Field(
        default={}, sa_column=Column(JSONB, nullable=False)
    )
    transformer_params: dict[str, Any] = Field(
        default={}, sa_column=Column(JSONB, nullable=False)
    )
    filter: Optional[Filter] = Field(
        default=None, sa_column=Column(JSONB, nullable=True)
    )


class Binding(BindingBase, table=True):
    __tablename__ = "bindings"
    binding_id: int = Field(default=None, primary_key=True)
    created_at: datetime = Field(
        default=None,
        sa_column=Column(
            DateTime(timezone=True), nullable=False, server_default=func.now()
        ),
    )
    updated_at: datetime = Field(
        default=None,
        sa_column=Column(
            DateTime(timezone=True),
            nullable=False,
            server_default=func.now(),
            onupdate=func.now(),
        ),
    )
    status: str = Field(default=BindingStatus.PENDING, nullable=False)
    collection_id: str = Field(
        sa_column_args=(ForeignKey("collections.collection_id", ondelete="CASCADE"),),
    )
    collection: "Collection" = Relationship(
        back_populates="bindings", sa_relationship_kwargs={"lazy": "selectin"}
    )
    transformer: Transformer = Relationship(
        back_populates="bindings", sa_relationship_kwargs={"lazy": "selectin"}
    )
    index: Index = Relationship(
        back_populates="bindings", sa_relationship_kwargs={"lazy": "selectin"}
    )

    def __repr__(self):
        return (
            f"<Binding("
            f"id={self.binding_id}, "
            f"status={self.status}, "
            f"collection='{self.collection_id}', "
            f"transformer='{self.transformer_id}', "
            f"index='{self.index_id}')>"
        )

    @property
    def filter_obj(self):
        if self.filter is not None:
            return Filter.model_validate(self.filter)
        return None


# Passed to the API from the client, can use name or id for collection
class BindingCreate(BindingBase):
    collection_name: Optional[str] = None
    collection_id: Optional[str] = None
    transformer_id: str
    index_id: str

    # Ensure either `_name` or `_id` is provided for collection
    @model_validator(mode="before")
    def check_identifiers(cls, values):
        if not values.get("collection_name") and not values.get("collection_id"):
            raise ValueError("Either collection_name or collection_id must be provided")
        return values


class BindingUpdate(BindingBase):
    description: Optional[str] = None
    execution_params: Optional[dict[str, Any]] = None
    transformer_params: Optional[dict[str, Any]] = None
    filter: Optional[Filter] = None
    status: Optional[str] = None


class BindingRead(BindingBase):
    binding_id: int
    created_at: datetime
    updated_at: datetime
    status: str
    collection: Collection
    transformer: Transformer
    index: Index
    filter: Optional[Filter]



from abc import ABC, abstractmethod


class StorageClient(ABC):
    """Abstract base class for storage clients."""

    @abstractmethod
    def is_authenticated(self) -> bool:
        pass

    @abstractmethod
    def list_buckets(self) -> list[str]:
        pass

    @abstractmethod
    def upload_object(
        self, fileobj, bucket_name: str, object_name: str, rewind: bool = True
    ) -> dict:
        pass

    @abstractmethod
    def generate_presigned_url(
        self, bucket_name: str, object_name: str, expiration: int = 3600
    ) -> str:
        pass

    @abstractmethod
    def delete_object(self, bucket_name: str, object_name: str) -> None:
        pass



from datetime import datetime
from typing import Any, Optional, TYPE_CHECKING

from sqlalchemy import Column, DateTime, JSON, func, String
from sqlmodel import SQLModel, Field, Relationship

if TYPE_CHECKING:
    from lexy.models.binding import Binding


def default_index_table_name(context) -> str:
    index_id = context.get_current_parameters()["index_id"]
    return f"zzidx__{index_id}"


class IndexBase(SQLModel):
    index_id: str = Field(
        default=None,
        primary_key=True,
        min_length=1,
        max_length=56,
        # Postgres identifiers are limited to 63 characters (so 63 - len('zzidx__') = 56)
        # TODO: switch back to `regex=` (or `pattern=`) once SQLModel bug is fixed
        #   https://github.com/tiangolo/sqlmodel/discussions/735
        schema_extra={"pattern": r"^[a-z_][a-z0-9_]{0,55}$"},
    )
    description: Optional[str] = None
    index_table_schema: Optional[dict[str, Any]] = Field(
        sa_column=Column(JSON, nullable=False), default={}
    )
    index_fields: Optional[dict[str, Any]] = Field(
        sa_column=Column(JSON, nullable=False), default={}
    )


class Index(IndexBase, table=True):
    __tablename__ = "indexes"
    created_at: datetime = Field(
        default=None,
        sa_column=Column(
            DateTime(timezone=True), nullable=False, server_default=func.now()
        ),
    )
    updated_at: datetime = Field(
        default=None,
        sa_column=Column(
            DateTime(timezone=True),
            nullable=False,
            server_default=func.now(),
            onupdate=func.now(),
        ),
    )
    bindings: list["Binding"] = Relationship(
        back_populates="index", sa_relationship_kwargs={"cascade": "all, delete-orphan"}
    )
    index_table_name: str = Field(
        default=None,
        # Postgres identifiers are limited to 63 characters
        # TODO: switch back to `regex=` (or `pattern=`) once SQLModel bug is fixed
        #   https://github.com/tiangolo/sqlmodel/discussions/735
        schema_extra={"pattern": r"^[a-z_][a-z0-9_]{0,62}$"},
        sa_column=Column(
            String, default=default_index_table_name, nullable=False, unique=True
        ),
    )


class IndexCreate(IndexBase):
    pass


class IndexUpdate(IndexBase):
    description: Optional[str] = None
    index_table_schema: Optional[dict[str, Any]] = None
    index_fields: Optional[dict[str, Any]] = None



import logging

from lexy.db.init_db import init_db

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def init() -> None:
    init_db(seed_data=True)


def main() -> None:
    logger.info("Creating initial seed data")
    init()
    logger.info("Initial seed data created")


if __name__ == "__main__":
    main()



import logging

from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.session import close_all_sessions
from sqlmodel import SQLModel

from lexy.core.config import settings as app_settings
from lexy.db.sample_data import default_data, sample_docs
from lexy.db.session import sync_engine
from lexy import models  # noqa: make sure all models are imported before initializing DB


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
SyncSessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=sync_engine)

db = SyncSessionLocal()


def add_default_data_to_db(session=db):
    # issue a warning if default seed data already exists in the database

    logger.info("Adding default collections")
    if session.query(models.Collection).count() > 0:
        logger.warning("Collection data already exists - skipping collection data")
    else:
        for c in default_data["collections"]:
            session.add(models.Collection(**c))
        session.commit()

    logger.info("Adding default transformers")
    if session.query(models.Transformer).count() > 0:
        logger.warning("Transformer data already exists - skipping transformer data")
    else:
        for t in default_data["transformers"]:
            session.add(models.Transformer(**t))
        session.commit()

    logger.info("Adding default indexes")
    if session.query(models.Index).count() > 0:
        logger.warning("Index data already exists - skipping index data")
    else:
        for i in default_data["indexes"]:
            session.add(models.Index(**i))
        session.commit()

    logger.info("Adding default bindings")
    if session.query(models.Binding).count() > 0:
        logger.warning("Binding data already exists - skipping binding data")
    else:
        for b in default_data["bindings"]:
            c = (
                session.query(models.Collection)
                .filter(models.Collection.collection_name == b["collection_name"])
                .first()
            )
            if c:
                session.add(models.Binding(**b, collection_id=c.collection_id))
            else:
                logger.warning(
                    f"Collection '{b['collection_name']}' not found for seed "
                    f"binding - skipping binding"
                )
        session.commit()


def add_sample_docs_to_db(session=db):
    # issue a warning if sample documents already exist in the database
    logger.info("Adding sample documents")
    if session.query(models.Document).count() > 0:
        logger.warning("Sample documents already exist - skipping sample documents")
    else:
        # adding sample documents for code collection only
        for doc in sample_docs["code_collection_sample_docs"]:
            c = (
                session.query(models.Collection)
                .filter(models.Collection.collection_name == doc["collection_name"])
                .first()
            )
            if c:
                session.add(models.Document(**doc, collection_id=c.collection_id))
                session.commit()
            else:
                logger.warning(
                    f"Collection '{doc['collection_name']}' not found for sample "
                    f"document - skipping sample document"
                )


def add_first_superuser_to_db(session=db, settings=app_settings):
    superuser = (
        session.query(models.User)
        .filter(models.User.email == settings.FIRST_SUPERUSER_EMAIL)
        .first()
    )
    if superuser:
        # issue a warning if superuser already exists in the database
        logger.warning("Superuser already exists - skipping superuser")
    else:
        # adding superuser
        session.add(
            models.User.create(
                email=settings.FIRST_SUPERUSER_EMAIL,
                password=settings.FIRST_SUPERUSER_PASSWORD.get_secret_value(),
                is_superuser=True,
            )
        )
        session.commit()


def init_db(session=db, seed_data=False):
    logger.info(f"Initializing database (seed_data={seed_data})")

    logger.info("Creating tables (SQLModel.metadata.create_all)")
    SQLModel.metadata.create_all(sync_engine)

    # TODO: add data via crud functions instead of directly adding to the session
    if seed_data is True:
        logger.info("Adding superuser")
        add_first_superuser_to_db(session)

        logger.info("Adding default seed data")
        add_default_data_to_db(session)

        logger.info("Adding sample documents")
        add_sample_docs_to_db(session)

    logger.info("Finished initializing database")


def drop_tables(session=db, drop_all=True, dry_run=True):
    """Drop all tables in the database

    WARNING: This will drop all indexes and celery tables too!
    If you get an error, you'll need to initiate IndexManager first.

    Args:
        session: database session
        drop_all: drop all tables, including indexes and celery tables
        dry_run: whether to actually drop the tables
    """
    logger.debug("Running drop_tables")
    session.commit()
    session.close_all()
    close_all_sessions()
    if drop_all:
        SQLModel.metadata.reflect(bind=session.bind)
    db_table_names = ", ".join(SQLModel.metadata.tables.keys())
    logger.info(
        f"Dropping database tables..."
        f"\n\tengine.url: {sync_engine.url}"
        f"\n\tdb_table_names: {db_table_names}"
    )
    if dry_run is False:
        SQLModel.metadata.drop_all(sync_engine)
        logger.info("Dropped tables")
    else:
        logger.info("Did not drop tables -- set `dry_run=False` to drop")


def reset_db(session=db, drop_all=True, dry_run=True):
    """Reset the database

    WARNING: this may not work as expected - use the make target `drop-db-tables`
    instead
    """
    logger.info("Resetting database")
    drop_tables(session, drop_all=drop_all, dry_run=dry_run)
    init_db(session, seed_data=True)
    logger.info("Finished resetting database")



"""Module for CRUD helpers."""

from sqlmodel import select
from sqlmodel.ext.asyncio.session import AsyncSession

from lexy.models import Collection, Document, Index, Transformer


async def get_collection_by_id(
    *, session: AsyncSession, collection_id: str
) -> Collection | None:
    """Get a collection by id."""
    result = await session.exec(
        select(Collection).where(Collection.collection_id == collection_id)
    )
    collection = result.first()
    return collection


async def get_collection_by_name(
    *, session: AsyncSession, collection_name: str
) -> Collection | None:
    """Get a collection by name."""
    result = await session.exec(
        select(Collection).where(Collection.collection_name == collection_name)
    )
    collection = result.first()
    return collection


async def get_document_by_id(
    *, session: AsyncSession, document_id: str
) -> Document | None:
    """Get a document by id."""
    result = await session.exec(
        select(Document).where(Document.document_id == document_id)
    )
    document = result.first()
    return document


async def get_documents_by_collection_id(
    *, session: AsyncSession, collection_id: str
) -> list[Document]:
    """Get all documents in a collection."""
    result = await session.exec(
        select(Document).where(Document.collection_id == collection_id)
    )
    documents = result.all()
    return documents


async def get_documents_by_collection_id_and_content(
    *, session: AsyncSession, collection_id: str, content: str
) -> list[Document]:
    """Get all documents in a collection with a specific value for `content`."""
    result = await session.exec(
        select(Document)
        .where(Document.collection_id == collection_id)
        .where(Document.content == content)
    )
    documents = result.all()
    return documents


async def get_index_by_id(*, session: AsyncSession, index_id: str) -> Index | None:
    """Get an index by id."""
    result = await session.exec(select(Index).where(Index.index_id == index_id))
    index = result.first()
    return index


async def get_transformer_by_id(
    *, session: AsyncSession, transformer_id: str
) -> Transformer | None:
    """Get a transformer by id."""
    result = await session.exec(
        select(Transformer).where(Transformer.transformer_id == transformer_id)
    )
    transformer = result.first()
    return transformer



from datetime import timedelta

from fastapi import APIRouter, Depends, HTTPException, status
from fastapi.security import OAuth2PasswordRequestForm
from sqlmodel.ext.asyncio.session import AsyncSession

from lexy.api import deps
from lexy.core.config import settings
from lexy.core.security import create_access_token
from lexy.schemas.token import Token


router = APIRouter()


@router.post("/login/access-token", response_model=Token)
async def login_for_access_token(
    form_data: OAuth2PasswordRequestForm = Depends(),
    db: AsyncSession = Depends(deps.get_db),
) -> Token:
    user = await deps.authenticate_user(
        db, email=form_data.username, password=form_data.password
    )
    if not user:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Incorrect username or password",
            headers={"WWW-Authenticate": "Bearer"},
        )
    access_token_expires = timedelta(minutes=settings.ACCESS_TOKEN_EXPIRE_MINUTES)
    access_token = create_access_token(
        data={"sub": user.email}, expires_delta=access_token_expires
    )
    return Token(access_token=access_token, token_type="bearer")



import logging

from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql import text

from lexy.db.session import sync_engine


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)
SyncSessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=sync_engine)


def init() -> None:
    try:
        db = SyncSessionLocal()
        # Try to create session to check if DB is awake
        db.execute(text("SELECT 1"))
    except Exception as e:
        logger.error(e)
        raise e


def main() -> None:
    logger.info("Attempting to connect to database")
    init()
    logger.info("Connected to database")


if __name__ == "__main__":
    main()



from fastapi import Depends, HTTPException, status
from jose import JWTError, jwt
from sqlalchemy import select
from sqlmodel.ext.asyncio.session import AsyncSession

from lexy.core.config import settings
from lexy.core.security import ALGORITHM, oauth2_scheme, verify_password
from lexy.db.session import get_session as get_db, get_sync_engine
from lexy.models.user import User
from lexy.schemas.token import TokenData
from lexy.indexes import IndexManager


async def get_user_by_email(db: AsyncSession, email: str) -> User:
    result = await db.execute(select(User).where(User.email == email))
    user = result.scalars().first()
    return user


async def get_current_user(
    db: AsyncSession = Depends(get_db), token: str = Depends(oauth2_scheme)
) -> User:
    credentials_exception = HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Could not validate credentials",
        headers={"WWW-Authenticate": "Bearer"},
    )
    try:
        payload = jwt.decode(
            token, settings.SECRET_KEY.get_secret_value(), algorithms=[ALGORITHM]
        )
        email: str = payload.get("sub")
        if email is None:
            raise credentials_exception
        token_data = TokenData(email=email)
    except JWTError:
        raise credentials_exception
    user = await get_user_by_email(db=db, email=token_data.email)
    if user is None:
        raise credentials_exception
    return user


async def get_current_active_user(current_user: User = Depends(get_current_user)):
    if not current_user.is_active:
        raise HTTPException(status_code=400, detail="Inactive user")
    return current_user


async def get_current_active_superuser(
    current_user: User = Depends(get_current_active_user),
):
    if not current_user.is_superuser:
        raise HTTPException(
            status_code=403, detail="User does not have enough privileges"
        )
    return current_user


async def authenticate_user(db: AsyncSession, *, email: str, password: str):
    user = await get_user_by_email(db=db, email=email)
    if not user:
        return False
    if not verify_password(password, user.hashed_password):
        return False
    return user


_index_manager: IndexManager | None = None


def get_index_manager() -> IndexManager:
    global _index_manager
    if _index_manager is None:
        engine = get_sync_engine()
        _index_manager = IndexManager(engine=engine)
    return _index_manager


index_manager = get_index_manager()



from datetime import datetime, timedelta, timezone

from fastapi.security import OAuth2PasswordBearer
from jose import jwt
from passlib.context import CryptContext

from lexy.core.config import settings


pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")
oauth2_scheme = OAuth2PasswordBearer(
    tokenUrl=f"{settings.API_PREFIX}/login/access-token"
)

ALGORITHM = "HS256"


def get_password_hash(password):
    return pwd_context.hash(password)


def verify_password(plain_password, hashed_password):
    return pwd_context.verify(plain_password, hashed_password)


def create_access_token(data: dict, expires_delta: timedelta | None = None):
    to_encode = data.copy()
    if expires_delta:
        expire = datetime.now(timezone.utc) + expires_delta
    else:
        expire = datetime.now(timezone.utc) + timedelta(minutes=15)
    to_encode.update({"exp": expire})
    encoded_jwt = jwt.encode(
        to_encode, settings.SECRET_KEY.get_secret_value(), algorithm=ALGORITHM
    )
    return encoded_jwt



import logging
from typing import Optional, TYPE_CHECKING

from fastapi import Depends

from lexy.storage.client import get_storage_client
from lexy.core.config import settings
from lexy.models import Binding, Document
from lexy.schemas.filters import filter_documents
from lexy.core.celery_tasks import celery, save_records_to_index

if TYPE_CHECKING:
    from lexy.storage.client import StorageClient


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


async def generate_tasks_for_document(
    doc: Document,
    storage_client: Optional["StorageClient"] = Depends(get_storage_client),
) -> list[dict]:
    """Generate tasks for a document based on bindings and filters.

    Args:
        doc (Document): The document for which to generate tasks
        storage_client (StorageClient): The storage client to use for generating
            presigned object URLs

    Returns:
        list[dict]: A list of tasks that were created for the document
    """
    logger.debug(f"Generating tasks for document {doc.document_id}")

    # initiate list of tasks
    tasks = []

    # loop through bindings for this document
    for binding in doc.collection.bindings:
        # check if binding is enabled
        if binding.status != "on":
            logger.debug(
                f"Skipping binding {binding} because it is not enabled (status: "
                f"{binding.status})"
            )
            continue

        # check if document matches binding filters
        if binding.filter and not binding.filter_obj.document_meets_conditions(doc):
            logger.debug(
                f"Skipping binding {binding} because document does not match filters"
            )
            continue

        # generate the task
        if doc.is_stored_object:
            if not storage_client:
                raise ValueError(
                    "Storage client must be configured when generating tasks for "
                    "stored objects"
                )
            doc.refresh_object_urls(storage_client=storage_client)
        task_name = binding.transformer.celery_task_name
        task = celery.send_task(
            task_name,
            args=[doc],
            kwargs=binding.transformer_params,
            link=save_records_to_index.s(
                document_id=doc.document_id,
                text=doc.content,
                index_id=binding.index_id,
                binding_id=binding.binding_id,
            ),
        )

        tasks.append({"task_id": task.id, "document_id": doc.document_id})

    logger.debug(
        f"Created {len(tasks)} tasks for document {doc.document_id}: "
        f"[{', '.join([t['task_id'] for t in tasks])}]"
    )
    return tasks


# TODO: remove session arg and run async after update to SQLAlchemy 2.0
def process_new_binding(
    session,
    binding: Binding,
    storage_client: Optional["StorageClient"] = Depends(get_storage_client),
) -> tuple[Binding, list[dict]]:
    """Process a new binding.

    Steps involved:
        1. Infer values for "lexy_index_fields" if not specified
        2. Create tasks for all documents in the collection that match the binding's
           filters
        3. Switch binding status to 'on'

    Args:
        binding (Binding): The binding to process
        storage_client (StorageClient): The storage client to use for generating
            presigned object URLs

    Returns:
        tuple[Binding, list[dict]]: The binding and the tasks that were created

    Raises:
        ValueError:
            - If the binding does not have a transformer or an index associated with it
            - If the binding does not have `lexy_index_fields` defined in
              `binding.transformer_params` AND `binding.index.index_fields` is not
              defined

    """
    logger.info(f"Processing new {binding = }")

    # check if binding has a valid transformer
    if binding.transformer is None:
        raise ValueError(f"{binding = } does not have a transformer associated with it")

    # check if binding has a valid index
    if binding.index is None:
        raise ValueError(f"{binding = } does not have an index associated with it")

    # check that the binding has lexy_index_fields defined
    if "lexy_index_fields" not in binding.transformer_params.keys():
        logger.info(
            f"{binding = } does not have 'lexy_index_fields' defined in "
            f"'transformer_params'"
        )
        if len(binding.index.index_fields) > 0:
            logger.info(
                f"Will assign them using 'index_fields' from index "
                f"'{binding.index.index_id}'."
            )

            # This next line is required because binding.transformer_params is a JSON
            # field. See https://github.com/sqlalchemy/sqlalchemy/discussions/6473
            binding.transformer_params = dict(binding.transformer_params)

            binding.transformer_params["lexy_index_fields"] = list(
                binding.index.index_fields.keys()
            )
            logger.info(
                f"Updated binding.transformer_params: {binding.transformer_params}"
            )
        else:
            raise ValueError(
                f"{binding = } does not have 'lexy_index_fields' defined in "
                f"'transformer_params' and index '{binding.index.index_id}' does not "
                f"have 'index_fields' defined. Either define 'lexy_index_fields' in "
                f"'transformer_params' or define 'index_fields' for index "
                f"'{binding.index.index_id}'."
            )

    # filter documents in the collection that match the binding filters
    if binding.filter:
        filter_obj = binding.filter_obj
        documents = filter_documents(binding.collection.documents, filter_obj)
    else:
        documents = binding.collection.documents

    # initiate list of tasks
    tasks = []

    # generate tasks for documents
    task_name = binding.transformer.celery_task_name
    for doc in documents:
        if doc.is_stored_object:
            if not storage_client:
                raise ValueError(
                    "Storage client must be configured when generating tasks for "
                    "stored objects"
                )
            doc.refresh_object_urls(storage_client=storage_client)
        task = celery.send_task(
            task_name,
            args=[doc],
            kwargs=binding.transformer_params,
            link=save_records_to_index.s(
                document_id=doc.document_id,
                text=doc.content,
                index_id=binding.index_id,
                binding_id=binding.binding_id,
            ),
        )
        tasks.append({"task_id": task.id, "document_id": doc.document_id})

    # switch binding status to 'on'
    prev_status = binding.status
    binding.status = "on"
    logger.info(f"Set status for {binding = } - from '{prev_status}' to 'on'")

    logger.info(
        f"Created {len(tasks)} tasks for {binding = }: "
        f"[{', '.join([t['task_id'] for t in tasks])}]"
    )

    return binding, tasks


def restart_celery_worker(workername: str) -> dict:
    """Restart a celery worker.

    Args:
        workername: The hostname of the worker to restart

    Returns:
        dict: The response from the worker
    """
    logger.debug(f"Broadcasting signal 'pool_restart' to celery worker {workername}")

    reload_modules = ["lexy.indexes", "lexy.core.celery_tasks"]
    transformer_modules = list(settings.worker_transformer_imports)
    reload_modules.extend(transformer_modules)
    print("reload_modules:", reload_modules)

    response = celery.control.broadcast(
        "pool_restart",
        arguments={"reload": True, "modules": reload_modules},
        destination=[workername],
        reply=True,
        timeout=3,
    )
    if response and "ok" in response[0][workername]:
        logger.info(f"Restarting '{workername}' worker's pool")
    else:
        logger.error(response)
        reason = error_reason(workername, response)
        logger.error(f"Failed to restart the '{workername}' pool: {reason}")
    return response


def error_reason(workername, response):
    """Extracts error message from response."""
    for res in response:
        try:
            return res[workername].get("error", "Unknown reason")
        except KeyError:
            pass
    logger.error("Failed to extract error reason from '%s'", response)
    return "Unknown reason"



from enum import Enum
from typing import Any, Iterable, TYPE_CHECKING

from pydantic import BaseModel, Field, ValidationInfo, field_validator

if TYPE_CHECKING:
    from lexy.models.document import Document


class Operation(str, Enum):
    EQUALS = "equals"
    EQUALS_CI = "equals_ci"  # Case-insensitive equals
    LESS_THAN = "less_than"
    LESS_THAN_OR_EQUALS = "less_than_or_equals"
    GREATER_THAN = "greater_than"
    GREATER_THAN_OR_EQUALS = "greater_than_or_equals"
    CONTAINS = "contains"
    CONTAINS_CI = "contains_ci"  # Case-insensitive contains
    STARTS_WITH = "starts_with"
    STARTS_WITH_CI = "starts_with_ci"  # Case-insensitive starts with
    ENDS_WITH = "ends_with"
    ENDS_WITH_CI = "ends_with_ci"  # Case-insensitive ends with
    IN = "in"


class FilterCondition(BaseModel):
    field: str = Field(..., description="The field to apply the filter on")
    operation: str = Field(..., description="The operation to perform")
    value: Any = Field(..., description="The value to compare against")
    negate: bool = Field(False, description="Whether to negate the filter condition")

    @field_validator("value")
    def validate_value(cls, v: Any, info: ValidationInfo) -> Any:
        operation = info.data.get("operation")

        # For numeric operations
        if operation in [
            Operation.LESS_THAN,
            Operation.LESS_THAN_OR_EQUALS,
            Operation.GREATER_THAN,
            Operation.GREATER_THAN_OR_EQUALS,
        ]:
            try:
                return float(v)
            except ValueError:
                raise ValueError(f"Value must be a number for operation '{operation}'")

        # For iterable operations
        if operation in [Operation.IN]:
            if not isinstance(v, Iterable):
                raise ValueError(
                    f"Value must be an iterable (list, tuple, set, or dict) for "
                    f"operation '{operation}'"
                )

        # For string operations
        if operation in [
            Operation.EQUALS_CI,
            Operation.CONTAINS,
            Operation.CONTAINS_CI,
            Operation.STARTS_WITH,
            Operation.STARTS_WITH_CI,
            Operation.ENDS_WITH,
            Operation.ENDS_WITH_CI,
        ]:
            if not isinstance(v, str):
                raise ValueError(f"Value must be a string for operation '{operation}'")

        return v


class Filter(BaseModel):
    conditions: list[FilterCondition]
    combination: str = Field("AND", description="Logical combination (AND/OR)")

    def document_meets_conditions(self, document: "Document") -> bool:
        if self.combination == "AND":
            return all(
                apply_filter_condition(document, condition)
                for condition in self.conditions
            )
        elif self.combination == "OR":
            return any(
                apply_filter_condition(document, condition)
                for condition in self.conditions
            )
        else:
            raise ValueError(f"Unsupported combination: {self.combination}")


# Function to apply individual filter condition to a document
def apply_filter_condition(document: "Document", condition: FilterCondition) -> bool:
    if condition.field.startswith("meta."):
        doc_value = document.meta.get(condition.field[5:], None)
    else:
        doc_value = getattr(document, condition.field, None)
    filter_value = condition.value

    if doc_value is None:
        # handle case where doc value is None
        if condition.operation in [Operation.EQUALS, Operation.EQUALS_CI]:
            result = filter_value is None
        elif condition.operation in [
            Operation.LESS_THAN,
            Operation.LESS_THAN_OR_EQUALS,
            Operation.GREATER_THAN,
            Operation.GREATER_THAN_OR_EQUALS,
            Operation.CONTAINS,
            Operation.CONTAINS_CI,
            Operation.STARTS_WITH,
            Operation.STARTS_WITH_CI,
            Operation.ENDS_WITH,
            Operation.ENDS_WITH_CI,
        ]:
            result = False
        elif condition.operation in [Operation.IN]:
            result = (
                None in filter_value if isinstance(filter_value, Iterable) else False
            )
        else:
            raise ValueError(
                f"Unsupported operation: {condition.operation} for doc value `None`"
            )
    else:
        # handle all other cases
        if condition.operation == Operation.EQUALS:
            result = doc_value == filter_value
        elif condition.operation == Operation.EQUALS_CI:
            result = str(doc_value).lower() == str(filter_value).lower()
        elif condition.operation == Operation.LESS_THAN:
            result = doc_value < filter_value
        elif condition.operation == Operation.LESS_THAN_OR_EQUALS:
            result = doc_value <= filter_value
        elif condition.operation == Operation.GREATER_THAN:
            result = doc_value > filter_value
        elif condition.operation == Operation.GREATER_THAN_OR_EQUALS:
            result = doc_value >= filter_value
        elif condition.operation == Operation.CONTAINS:
            result = filter_value in doc_value
        elif condition.operation == Operation.CONTAINS_CI:
            result = str(filter_value).lower() in str(doc_value).lower()
        elif condition.operation == Operation.STARTS_WITH:
            result = doc_value.startswith(filter_value)
        elif condition.operation == Operation.STARTS_WITH_CI:
            result = str(doc_value).lower().startswith(str(filter_value).lower())
        elif condition.operation == Operation.ENDS_WITH:
            result = doc_value.endswith(filter_value)
        elif condition.operation == Operation.ENDS_WITH_CI:
            result = str(doc_value).lower().endswith(str(filter_value).lower())
        elif condition.operation == Operation.IN:
            result = doc_value in filter_value
        else:
            raise ValueError(f"Unsupported operation: {condition.operation}")

    return not result if condition.negate else result


# Function to apply all filters to a list of documents
def filter_documents(
    docs: Iterable["Document"], filter_obj: Filter
) -> Iterable["Document"]:
    """Filter documents according to a filter object.

    Args:
        docs (Iterable[Document]): The documents to filter.
        filter_obj (Filter): The filter object to apply.

    Yields:
        Iterable[Document]: The filtered documents.

    Examples:
        >>> filter_obj = Filter(
        ...     conditions=[
        ...         FilterCondition(field='text', operation='contains', value='foo'),
        ...         FilterCondition(field='meta.bar', operation='equals', value='baz')
        ...     ],
        ...     combination='AND'
        ... )
        >>> docs = [
        ...     Document(content='foo bar', meta={'bar': 'baz'}),
        ...     Document(content='foo', meta={'bar': 'qux'})
        ... ]
        >>> list(filter_documents(docs, filter_obj))
        [Document(content='foo bar', meta={'bar': 'baz'})]
    """
    for doc in docs:
        if filter_obj.combination == "AND":
            if all(
                apply_filter_condition(doc, condition)
                for condition in filter_obj.conditions
            ):
                yield doc
        elif filter_obj.combination == "OR":
            if any(
                apply_filter_condition(doc, condition)
                for condition in filter_obj.conditions
            ):
                yield doc
        else:
            raise ValueError(f"Unsupported combination: {filter_obj.combination}")


def document_passes_filter(document: "Document", filter_obj: Filter) -> bool:
    if filter_obj.combination == "AND":
        return all(
            apply_filter_condition(document, condition)
            for condition in filter_obj.conditions
        )
    elif filter_obj.combination == "OR":
        return any(
            apply_filter_condition(document, condition)
            for condition in filter_obj.conditions
        )
    else:
        raise ValueError(f"Unsupported combination: {filter_obj.combination}")



"""Module for creating and managing indexes"""

import logging
from datetime import datetime, date, time
from typing import Dict, ForwardRef, Literal, Optional, Type
from uuid import UUID, uuid1, uuid3, uuid4, uuid5

from pydantic import create_model
from pgvector.sqlalchemy import Vector
from sqlalchemy import Column, DDL, event, inspect
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import sessionmaker
from sqlalchemy.engine import Engine, Inspector
from sqlmodel import SQLModel, Field, ForeignKey, Session, select
from sqlmodel.main import RelationshipInfo

from lexy.models.index import Index
from lexy.models.index_record import IndexRecordBaseTable


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

LEXY_INDEX_FIELD_TYPES: Dict = {
    "int": int,
    "integer": int,
    "float": float,
    "str": str,
    "string": str,
    "text": str,
    "bytes": bytes,
    "bytearray": bytearray,
    "date": date,
    "datetime": datetime,
    "time": time,
    "uuid": UUID,
    "uuid1": uuid1,
    "uuid3": uuid3,
    "uuid4": uuid4,
    "uuid5": uuid5,
    "bool": bool,
    "boolean": bool,
    "dict": dict,
    "object": dict,
    "list": list,
    "array": list,
}


# Mapping from tablename to class
def map_tablename_to_class(SQLModel) -> dict:
    tbl_to_cls = {}
    for mapper in SQLModel._sa_registry.mappers:
        cls = mapper.class_
        classname = cls.__name__
        if not classname.startswith("_"):
            tblname = cls.__tablename__
            tbl_to_cls[tblname] = cls
    return tbl_to_cls


TBLNAME_TO_CLASS = map_tablename_to_class(SQLModel)


# TODO: add hnsw index type (e.g., cosine, euclidean, etc.) to DDL statement
class IndexManager(object):
    TBLNAME_TO_CLASS: dict | None = TBLNAME_TO_CLASS

    def __init__(self, engine: Engine):
        self._engine: Engine = engine
        self._db: Session | None = None
        self._inspector: Inspector | None = None

        self.index_models: dict[str, Type[IndexRecordBaseTable]] = {}
        logger.info(
            f"IndexManager initialized with db url '{repr(self.db.bind.engine.url)}'"
        )

    @property
    def db(self) -> Session:
        if self._db is None:
            sync_session_local = sessionmaker(
                class_=Session, autocommit=False, autoflush=False, bind=self._engine
            )
            self._db = sync_session_local()
        return self._db

    @property
    def inspector(self) -> Inspector:
        if self._inspector is None:
            self._inspector = inspect(self.db.bind.engine)
        return self._inspector

    def get_indexes(self) -> list[Index]:
        """Get all indexes"""
        indexes = self.db.exec(select(Index)).all()
        return indexes

    def get_index(self, index_id: str) -> Index:
        """Get an index by id

        Args:
            index_id (str): index id

        Returns:
            Index: index

        Raises:
            ValueError: if index is not found
        """
        index = self.db.exec(select(Index).where(Index.index_id == index_id)).first()
        if not index:
            raise ValueError(f"Index {index_id} not found")
        return index

    def create_index_models(self):
        logger.info("Creating index models")
        indexes = self.get_indexes()
        for index in indexes:
            if self.index_models.get(index.index_id):
                logger.warning(
                    f"create_index_models -- Index model {index.index_id} already "
                    f"exists."
                )
            index_model = self.create_index_model(index)
            if self.table_exists(index.index_table_name):
                logger.info(
                    f"create_index_models -- Index table {index.index_table_name} "
                    f"already exists."
                )
            else:
                logger.info(
                    f"create_index_models -- Creating new Index table "
                    f"{index.index_table_name}."
                )
                created = self._create_index_table(index_model)
                if created:
                    logger.info(
                        f"create_index_models -- Created new Index table "
                        f"{index.index_table_name}."
                    )
                else:
                    logger.warning(
                        f"create_index_models -- Failed to create Index table "
                        f"{index.index_table_name}."
                    )

    def create_index_model(self, index: Index) -> Type[IndexRecordBaseTable]:
        index_table_name = index.index_table_name
        model_name = index.index_table_schema.get("title", index_table_name)

        # If index_table_name is already in SQLModel's _sa_registry, log a warning and
        # return the existing model
        if index_table_name in self.TBLNAME_TO_CLASS.keys():
            logger.warning(
                f"create_index_model -- Index table {index_table_name} exists in "
                f"TBLNAME_TO_CLASS.keys()."
            )
            index_model = self.TBLNAME_TO_CLASS.get(index_table_name)
            self.index_models[index.index_id] = index_model
            return index_model

        # Index field definitions
        field_defs = self.get_field_definitions(index.index_fields)

        # Set class vars and redefine foreign keys with the constraint
        # `ondelete='CASCADE'`
        class IndexRecordTable(IndexRecordBaseTable):
            __tablename__ = index.index_table_name
            __index_id__ = index.index_id
            document_id: Optional[UUID] = Field(
                sa_column_args=(
                    ForeignKey("documents.document_id", ondelete="CASCADE"),
                ),
                index=True,
                nullable=True,
            )
            binding_id: Optional[int] = Field(
                sa_column_args=(ForeignKey("bindings.binding_id", ondelete="CASCADE"),),
                index=True,
                nullable=True,
            )

        # Relationship definitions
        relationship_defs = {
            # Add the Document relationship, i.e., the equivalent of the field:
            #   document: "Document" = Relationship(back_populates="index_records")
            # TODO: add back_populates to RelationshipInfo() once index_records is
            #  added to Document
            "document": (ForwardRef("Document"), RelationshipInfo()),
            # Add the Binding relationship, i.e., the equivalent of the field:
            #   binding: "Binding" = Relationship(back_populates="index_records")
            # TODO: add back_populates to RelationshipInfo() once index_records is
            #  added to Binding
            "binding": (ForwardRef("Binding"), RelationshipInfo()),
        }

        index_model = create_model(
            model_name,
            __base__=(IndexRecordTable,),
            __cls_kwargs__={"table": True},
            __module__=__name__,
            **field_defs,
            **relationship_defs,
        )

        # Attach event listener to create index on embedding fields
        embedding_ddl = self.get_ddl_for_embedding_fields(
            index.index_fields, index_table_name
        )
        for colname, ddl in embedding_ddl.items():
            event.listen(
                index_model.__table__,
                "after_create",
                ddl,
            )

        self.index_models[index.index_id] = index_model
        return index_model

    def get_or_create_index_model(
        self, index: Index
    ) -> tuple[Type[IndexRecordBaseTable], bool]:
        """Get or create an index model from an Index object.

        Args:
            index (Index): The index object to get or create the model for.

        Returns:
            tuple[Type[IndexRecordBaseTable], bool]: A tuple of the index model and a
                boolean indicating whether the model was created.
        """
        if index.index_id in self.index_models.keys():
            logger.debug(
                f"get_or_create_index_model -- Index model {index.index_id} already "
                f"exists."
            )
            return self.index_models.get(index.index_id), False
        else:
            logger.debug(
                f"get_or_create_index_model -- Index model {index.index_id} does not "
                f"exist - will try to create it"
            )
            return self.create_index_model(index), True

    def create_index_model_and_table(self, index_id: str) -> tuple[bool, bool]:
        """Create a new index model and its associated table.

        **This requires that the index row already exists in the database.**

        Steps involved:
            1. Get the index object from the indexes table
            2. Create (or get) the corresponding index model
            3. Create associated index table in the database

        Args:
            index_id (str): The ID of the index to create

        Returns:
            tuple[bool, bool]: A tuple of booleans indicating whether the index model
                and table were created
        """
        logger.info(f"Creating new index model and table for {index_id = }")

        # Get the index object
        index = self.get_index(index_id)
        if index is None:
            raise ValueError(f"Index {index_id} not found")

        # Create (or get) the corresponding index model
        index_model, model_created = self.get_or_create_index_model(index)
        if not model_created:
            # NOTE: This should only happen if an index with the same name was
            #   previously created, and then deleted with option `drop_table=False`.
            #   To force the recreation of the index or table, the user should recreate
            #   the index, then delete it with option drop_table=True.
            logger.warning(f"Index model {index_id} already exists. Skipping creation.")

        # Create index table in the database
        table_created = self._create_index_table(index_model)
        if not table_created:
            logger.warning(
                f"Index table {index.index_table_name} already exists. Skipping "
                f"creation."
            )

        return model_created, table_created

    def drop_index_table(self, index_id: str) -> bool:
        """Drops the index table from the database and removes the index model from the
        index manager.

        Args:
            index_id (str): The ID of the index to drop.

        Returns:
            bool: True if the index table was dropped, False otherwise.
        """
        logger.info(f"Dropping index table for {index_id = }")

        index = self.get_index(index_id)

        # Pop to remove the index model from self.index_models
        index_model = self.index_models.pop(index_id, None)
        # If the index model doesn't exist, log a warning and return
        if index_model is None:
            logger.warning(f"Index model for '{index_id}' does not exist")
            return False

        # If the table doesn't exist, log a warning and return
        if not self.table_exists(index.index_table_name):
            logger.warning(f"Index table '{index.index_table_name}' does not exist")
            return False

        logger.debug(f"about to drop table for index_model: {index_model}")
        index_model.metadata.drop_all(
            self.db.bind.engine, tables=[index_model.__table__]
        )
        index_model.metadata.remove(index_model.__table__)
        # Clear the inspector cache to ensure that `self.table_exists` works as expected
        self.inspector.clear_cache()
        return True

    def table_exists(self, index_table_name: str, use_cache: bool = True) -> bool:
        if use_cache is not True:
            self.inspector.clear_cache()
        return self.inspector.has_table(index_table_name)

    def _create_index_table(self, index_model: Type[IndexRecordBaseTable]) -> bool:
        """Create index table from an Index model"""
        # If the tablename is None, raise an error
        if index_model.__tablename__ is None:
            raise ValueError("Index model must have a __tablename__ attribute")
        # If the table already exists, log a warning and return
        if self.table_exists(index_model.__tablename__):
            logger.warning(
                f"Index table {index_model.__tablename__} already exists. Skipping "
                f"creation."
            )
            return False
        index_model.metadata.create_all(self.db.bind.engine)
        # Clear the inspector cache to ensure that `self.table_exists` works as expected
        self.inspector.clear_cache()
        return True

    @staticmethod
    def get_field_definitions(index_fields: dict) -> dict[str, tuple[type, Field]]:
        """Get field definitions from index fields

        Args:
            index_fields: dict representation of index fields

        Returns:
            dict: field definitions

        Examples:
            >>> sample_fields_as_dict = {
            >>>      "embedding": {"type": "embedding", "extras": {"dims": 384, "model": "text.embeddings.minilm"}},
            >>>      "text": {"type": "string", "optional": True},
            >>> }
            >>> IndexManager.get_field_definitions(sample_fields_as_dict)

        """
        field_defs = {}
        for fk, fv in index_fields.items():
            fd_info = None

            if fv["type"] == "embedding":
                embedding_args = fv["extras"]
                fd_type = list[float]
                fd_info = Field(sa_column=Column(Vector(dim=embedding_args["dims"])))
            else:
                fd_type = LEXY_INDEX_FIELD_TYPES.get(fv["type"])
                if "optional" in fv and fv["optional"] is True:
                    fd_type = Optional[fd_type]

                if fv["type"] in ["dict", "object", "list", "array"]:
                    fd_info = Field(sa_column=Column(JSONB), default=None)

            field_defs[fk] = (fd_type, fd_info)

        return field_defs

    @staticmethod
    def get_ddl_for_embedding_fields(
        index_fields: dict,
        tablename: str,
        vector_ops_type: Literal[
            "vector_cosine_ops", "vector_l2_ops", "vector_ip_ops"
        ] = "vector_cosine_ops",
    ) -> dict:
        """Get DDL statements for index fields of type 'embedding'

        Args:
            index_fields: dict representation of index fields
            tablename: Name of index table
            vector_ops_type: Type of vector operations to use for the index. Default is
                'vector_cosine_ops'.

        Returns:
            dict[str, DDL]: ddl statements

        Examples:
            >>> index_fields_as_dict = {
            >>>      "embedding": {"type": "embedding", "extras": {"dims": 384, "model": "text.embeddings.minilm"}},
            >>>      "text": {"type": "string", "optional": True},
            >>> }
            >>> IndexManager.get_ddl_for_embedding_fields(index_fields_as_dict, 'my_index_table')

        """
        embedding_ddl = {}
        for fk, fv in index_fields.items():
            if fv["type"] == "embedding":
                colname = fk
                embedding_ddl[fk] = DDL(
                    f"CREATE INDEX IF NOT EXISTS ix_{tablename}_{colname}_hnsw "
                    f"ON {tablename} "
                    f"USING hnsw ({colname} {vector_ops_type}) "
                    f"WITH ("
                    f"  m = 16, "
                    f"  ef_construction = 64 "
                    f");"
                )
        return embedding_ddl
